/* --------------------------------------------------------------------------
 * DatUtil v1.3 (19 May 2000)
 * Written by Logiqx (logiqx@emuhq.com)
 *
 * A simple little utility for converting, cleansing and comparing ROM manager
 * dats (e.g. ClrMamePro, RomCenter). Also supports ROM lists generated by
 * emulators such as Retrocade, Callus and System 16 so that dats can be checked
 * for missing/spurious ROMs.
 *
 * Additional formats can easily be added if you have know a little C.
 * -------------------------------------------------------------------------- */

/* --- The standard includes --- */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* --- My type definitions and macros --- */

#include "datutil.h"


/* --- The Dat formats and ROM list formats supported --- */

LOAD_SAVE_FORMAT(romcenter_dat, "RomCenter Data", "ROMCENTER", ROMCENT_SUPPORT)
LOAD_SAVE_FORMAT(mame_listinfo, "MAME Listinfo", "LISTINFO", FULL_SUPPORT)
LOAD_SAVE_FORMAT(romcenter2_dat, "RomCenter 2 Data", "ROMCENTER2", ROMCENT2_SUPPORT)
LOAD_SAVE_FORMAT(mess_listinfo, "MESS Listinfo", "LISTINFO", FULL_SUPPORT)
LOAD_ONLY_FORMAT(raine_gameinfo, "RAINE Gameinfo", FULL_SUPPORT)
LOAD_ONLY_FORMAT(shark_listinfo, "Shark Listinfo", SIZE_SUPPORT|ALT_ROM_SUPPORT)
LOAD_ONLY_FORMAT(callus_gamelist, "Callus Gamelist", PARENT_SUPPORT)
LOAD_ONLY_FORMAT(retrocade_romlist, "Retrocade Romlist", 0)
LOAD_ONLY_FORMAT(system16_gcs, "System16 GCS", PARENT_SUPPORT|ONE_GAME_ONLY|SAMPLE_SUPPORT)
LOAD_ONLY_FORMAT(jff_txt, "Just For Fun Text", SIZE_SUPPORT)
LOAD_ONLY_FORMAT(sparcade_romlist, "Sparcade Romlist", ALT_ROM_SUPPORT)
LOAD_ONLY_FORMAT(sparcade_rat, "Sparcade RAT File", ALT_ROM_SUPPORT)
LOAD_ONLY_FORMAT(jas_romlist, "JAS Romlist", SIZE_SUPPORT)

const struct DatFormat *formats[] =
{
	FORMAT(romcenter_dat)
	FORMAT(mame_listinfo)
	FORMAT(romcenter2_dat)
	FORMAT(mess_listinfo)
	FORMAT(raine_gameinfo)
	FORMAT(shark_listinfo)
	FORMAT(callus_gamelist)
	FORMAT(retrocade_romlist)
	FORMAT(system16_gcs)
	FORMAT(jff_txt)
	/* FORMAT(sparcade_romlist) - Not enabled. See reasons in source */
	FORMAT(sparcade_rat)
	FORMAT(jas_romlist)
	0
};

/* --------------------------------------------------------------------------
 * The main() function just works out what the user wants to do then calls the
 * appropriate function (convert_dat or compare_dats).
 *
 * Uses the getopt() function from unistd to interpret command line options.
 * -------------------------------------------------------------------------- */

int main(int argc, char **argv)
{
	/* --- For getopt function --- */

	extern char *optarg;
	extern int optind;

	/* --- Options chosen by user ---*/

	char p_game[GAME_NAME_LENGTH+1]="";
	char p_outfile[LONG_NAME_LENGTH+1]="";
	char p_outmode[2]="";
	char p_format[10]="";
	int p_clones = 0;
	int p_sort=0;

	/* --- Misc variables ---*/

	int c, errflg = 0;

	/* --- Credits! --- */

	printf("===============================================================================\n");
#ifdef LONG_NAMES
	printf("DatUtil v1.3 (19 May 2000) - long name version\n");
#else
	printf("DatUtil v1.3 (19 May 2000)\n");
#endif
	printf("Written by Logiqx (logiqx@emuhq.com)\n");
	printf("===============================================================================\n");

	/* --- Get the options specified on the command line --- */

	while ((c = getopt(argc, argv, "g:co:a:f:s")) != EOF)
	switch (c)
	{
		case 'g':
			strcpy(p_game, optarg);							/* Game to extract */
			break;
		case 'c':
			p_clones++;													/* ...and it's clones */
			break;
		case 'o':
			strcpy(p_outfile, optarg);					/* Output file - overwrite */
			if (p_outmode[0])
				errflg++;
			else
				sprintf(p_outmode, "w");
			break;
		case 'a':
			strcpy(p_outfile, optarg);					/* Output file - append */
			if (p_outmode[0])
				errflg++;
			else
				sprintf(p_outmode, "a");
			break;
		case 'f':
			strcpy(p_format, optarg);						/* Output file format */
			UPPER(p_format);
			break;
		case 's':
			p_sort++;														/* Sort */
			break;
		case '?':
			errflg++;														/* User wants help! */
	}

	/* --- The user must specify either one or two dat files --- */

	if (argc-optind<1 || argc-optind>2)
		errflg++;

	/* --- Display the help page if required --- */

	if (errflg)
	{
		printf("Usage 1: datutil [-g<game> [-c]] [-o|a<outfile> [-f<format>]] [-s] <datfile>\n\n");
		printf("Converts datfile to a different format (the output format can be specified by\nusing -f). To specify the output filename use the -o option (or -a for append).\nIndividual games can be converted using the -g option (using -c will also\nconvert it's clones). If required, the -s option can be used to sort the games.\n\n");
		printf("Usage 2: datutil [-g<game> [-c]] [-o|a<logfile>] <datfile1> <datfile2>\n\n");
		printf("Compares two datfiles and reports any differences. To specify the name for the\nlogfile use the -o option (or -a for append). Individual games can be compared\nusing the -g option (using -c will also also compare it's clones).\n\n");
		exit (1);
	}

	/* --- If the user didn't specify the outmode, use overwrite --- */

	if (!p_outmode[0])
		sprintf(p_outmode, "w");

	/* --- Depending on the number of files, either convert or compare --- */

	if (optind==argc-1)
	{
		if (p_outfile[0]=='\0')
			sprintf(p_outfile, "datutil.dat");
		errflg=convert_dat(argv[optind], &p_game, p_clones, &p_outfile, &p_outmode, &p_format, p_sort);
	}
	else
	{
		if (p_outfile[0]=='\0')
			sprintf(p_outfile, "datutil.log");
		errflg=compare_dats(argv[optind], argv[optind+1], &p_game, p_clones, &p_outfile, &p_outmode);
	}

	/* --- All done --- */

	exit(errflg);
}

/* --------------------------------------------------------------------------
 * Dat conversion / extraction of individual game (and clones if required).
 * Basically just identifies the source dat, loads it into memory then
 * saves it back out again in the appropraiet format.
 * -------------------------------------------------------------------------- */

int convert_dat(char *datname, char *game, int clones, char *outfile, char *outmode, char *format, int sort)
{
	struct dat dat1;
	int errflg = 0;
	int i, save_format=-1;

	init_dat(&dat1, datname);

	/* --- If dat was identified and contains sufficient details to save... --- */

	if (dat1.flags & SAVE_SUPPORT)
	{
		/* --- Load the dat into memory --- */

		LOAD_DAT(dat1)

		/* --- If successful --- */

		if (!errflg)
		{
			display_dat_info(&dat1);

			if (sort)
				qsort(dat1.games, dat1.num_games, sizeof(struct game), parent_sort_function);

			/* --- Search for first save format different to original --- */
			for (i=0;formats[i];i++)
				if (save_format==-1 && formats[i]->save!=dat1.save)
					save_format=i;

			/* --- Overide with original save format if saving individual game --- */
			if (game[0]!='\0')
				for (i=0;formats[i];i++)
					if (formats[i]->save==dat1.save)
						save_format=i;

			/* --- Again, overide if format parameter was specified --- */
			if (format[0]!='\0')
				for (i=0;formats[i];i++)
					if (!strcmp(format, formats[i]->option))
						save_format=i;

			/* --- Now we've decided the format, save the dat --- */
			(formats[save_format]->save)(&dat1, outfile, outmode);
		}

		release_dat(&dat1);

		return(0);
	}
	else if (dat1.flags)
	{
		printf("%s does not contain sufficient information for conversion.\n", dat1.descr);
		return(1);
	}
}

/* --------------------------------------------------------------------------
 * Dat comparison. There are two algorithms here... both sort the games and
 * compare the resultant games lists, the difference is how ROMs are compared:
 *
 * 1) If neither dat is a Shark listinfo (which allows several names for each
 *    ROM) it sorts the ROMs and compares them in the same way as with games.
 * 2) If either dat is Shark listinfo then for each ROM in a game, every
 *    possible name is compared against all names for the ROMs in the other dat!
 * -------------------------------------------------------------------------- */

/* --- The maximum number of ROMs for one game --- */

#define MAX_GAME_ROMS 256

int compare_dats(char *datname1, char *datname2, char *game, int clones, char *outfile, char *outmode)
{
	struct dat dat1, dat2;
	struct game *games1, *games2;

	FILE *out;

	struct rom *rom_ptr;
	struct alt_rom *alt_rom_ptr;
	struct rom roms1[MAX_GAME_ROMS], roms2[MAX_GAME_ROMS];
	struct alt_rom alt_roms1[MAX_GAME_ROMS], alt_roms2[MAX_GAME_ROMS];
	int found, sizediff;
	char size1[SIZE_LENGTH+1], size2[SIZE_LENGTH+1];
	int num_roms1, num_roms2;

	int errflg = 0;

	int i=0, j=0, gamediff;
	int k, l, m, n, romdiff;
	int total_diffs=0;
	int bad_size1=0, bad_size2=0;
	int bad_romof1=0, bad_romof2=0;
	int bad_merge1=0, bad_merge2=0;
	int warning=0;

	/* --- Identify the dat types... --- */

	init_dat(&dat1, datname1);
	init_dat(&dat2, datname2);

	/* --- Load them in, display their summary and sort the games --- */

	if (dat1.flags && dat2.flags)
		LOAD_DAT(dat1)
	else
		errflg++;

	if (!errflg)
	{
		display_dat_info(&dat1);
		qsort(dat1.games, dat1.num_games, sizeof(struct game), game_sort_function);
	}

	if (!errflg && dat2.flags)
		LOAD_DAT(dat2)
	else
		errflg++;

	if (!errflg)
	{
		display_dat_info(&dat2);
		qsort(dat2.games, dat2.num_games, sizeof(struct game), game_sort_function);
	}
	
	/* --- Open the log file --- */

	if (!errflg && !(out=fopen(outfile, outmode)))
	{
		printf("Error opening %s\n", outfile);
		return(1);
	}

	/* --- If neither dat uses alternative ROM names, use sort method 1 --- */

	if (!errflg && !(dat1.flags & ALT_ROM_SUPPORT | dat2.flags & ALT_ROM_SUPPORT))
	{
		games1=dat1.games; games2=dat2.games;

		/* --- For all games --- */

		while (i<dat1.num_games || j<dat2.num_games)
		{
			/* --- If reached the end of either dat there is a game missing --- */

			if (i>=dat1.num_games) gamediff=1;
			if (j>=dat2.num_games) gamediff=-1;

			/* --- Else compare the game names --- */

			if (i<dat1.num_games && j<dat2.num_games)
				gamediff=strcmp(games1[i].name, games2[j].name);

			/* --- If the game names are the same then compare their details --- */

			if (gamediff==0)
			{
				/* --- Compare their parent details --- */

				if ((dat1.flags & PARENT_SUPPORT) && (dat2.flags & PARENT_SUPPORT) &&
					strcmp(games1[i].parent_name, games2[j].parent_name))
				{
					fprintf(out, "%s: **game** has different parents ('%s' and '%s')\n", games1[i].name, games1[i].parent_name, games2[j].parent_name);
					total_diffs++;
				}

				/* --- Compare their romof details --- */

				if ((dat1.flags & ROMOF_SUPPORT) && (dat2.flags & ROMOF_SUPPORT) &&
					strcmp(games1[i].romof, games2[j].romof))
				{
					fprintf(out, "%s: **game** has different romof values ('%s' and '%s')\n", games1[i].name, games1[i].romof, games2[j].romof);
					total_diffs++;
				}

				if ((dat1.flags & ROMOF_SUPPORT) && strcmp(games1[i].parent_name, games1[i].romof) &&
						strcmp(games1[i].romof, "neogeo") && !(dat2.flags & ROMOF_SUPPORT))
				{
					/*printf("%s %s\n", games1[i].parent_name, games1[i].romof);*/
					bad_romof2++;
				}

				if ((dat2.flags & ROMOF_SUPPORT) && strcmp(games2[j].parent_name, games2[j].romof) &&
						strcmp(games2[j].romof, "neogeo") && !(dat1.flags & ROMOF_SUPPORT))
				{
					/*printf("%s %s\n", games2[j].parent_name, games2[j].romof);*/
					bad_romof1++;
				}

				/* --- Load the ROMs from dat1 into buffer and sort --- */

				num_roms1=0;
				rom_ptr=games1[i].first_rom;
				while (rom_ptr)
				{
					strcpy(roms1[num_roms1].name, rom_ptr->name);
					strcpy(roms1[num_roms1].merge, rom_ptr->merge);
					strcpy(roms1[num_roms1].size, rom_ptr->size);
					strcpy(roms1[num_roms1].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;
					num_roms1++;
				}
				qsort(roms1, num_roms1, sizeof(struct rom), rom_sort_function);

				/* --- Load the ROMs from dat2 into buffer and sort --- */

				num_roms2=0;
				rom_ptr=games2[j].first_rom;
				while (rom_ptr)
				{
					strcpy(roms2[num_roms2].name, rom_ptr->name);
					strcpy(roms2[num_roms2].merge, rom_ptr->merge);
					strcpy(roms2[num_roms2].size, rom_ptr->size);
					strcpy(roms2[num_roms2].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;
					num_roms2++;
				}
				qsort(roms2, num_roms2, sizeof(struct rom), rom_sort_function);

				/* --- Compare the ROMs in buffers --- */

				k=l=0;

				while (k<num_roms1 || l<num_roms2 )
				{
					/* --- If reached the end of either, there are ROMs missing --- */

					if (k>=num_roms1) romdiff=1;
					if (l>=num_roms2) romdiff=-1;

					/* --- Else compare the ROM names --- */

					if (k<num_roms1 && l<num_roms2)
						romdiff=strcmp(roms1[k].name, roms2[l].name);

					/* --- If the names match, compare the ROM details --- */

					if (!romdiff)
					{
						/* --- Compare the merge names --- */

						if ((dat1.flags & MERGE_SUPPORT) && (dat2.flags & MERGE_SUPPORT) &&
							strcmp(roms1[k].merge, roms2[l].merge))
						{
							fprintf(out, "%s: %s has different merge names (%s - %s)\n", games1[i].name, roms1[k].name, roms1[k].merge, roms2[l].merge);
							total_diffs++;
						}

						if ((dat1.flags & MERGE_SUPPORT) && roms1[k].merge[0] && strcmp(roms1[k].name, roms1[k].merge) &&
								!(dat2.flags & MERGE_SUPPORT))
							bad_merge2++;

						if ((dat2.flags & MERGE_SUPPORT) && roms2[l].merge[0] && strcmp(roms2[l].name, roms2[l].merge) &&
								!(dat1.flags & MERGE_SUPPORT))
							bad_merge1++;

						/* --- Compare the sizes --- */

						if (roms1[k].size[0]!='\0' && roms2[l].size[0]!='\0' &&
							strcmp(roms1[k].size, roms2[l].size))
						{
							fprintf(out, "%s: %s has different sizes (%s - %s)\n", games1[i].name, roms1[k].name, roms1[k].size, roms2[l].size);
							total_diffs++;
						}

						if ((dat1.flags & dat2.flags & SIZE_SUPPORT) && roms1[k].size[0]=='\0')
							bad_size1++;

						if ((dat2.flags & dat2.flags & SIZE_SUPPORT) && roms2[l].size[0]=='\0')
							bad_size2++;

						/* --- Compare the CRCs --- */

						if ((dat1.flags & CRC_SUPPORT) && (dat2.flags & CRC_SUPPORT) &&
							strcmp(roms1[k].crc, roms2[l].crc))
						{
							fprintf(out, "%s: %s has different CRCs (%s - %s)\n", games1[i].name, roms1[k].name, roms1[k].crc, roms2[l].crc);
							total_diffs++;
						}

						/* --- Move on to next ROMs in buffers --- */

						k++; l++;
					}

					/* --- If ROM was missing in dat2 --- */

					if (romdiff<0)
					{
						if (dat1.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games1[i].name, roms1[k].name, roms1[k].crc, dat2.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games1[i].name, roms1[k].name, dat2.name);
						k++; total_diffs++;
					}

					/* --- If ROM was missing in dat1 --- */

					if (romdiff>0)
					{
						if (dat2.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games2[j].name, roms2[l].name, roms2[l].crc, dat1.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games2[j].name, roms2[l].name, dat1.name);
						l++; total_diffs++;
					}
				}

				/* --- Move on to next games --- */

				i++; j++;
			}

			/* --- If game was missing in dat2 --- */

			if (gamediff<0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games1[i].name, dat2.name);
				i++; total_diffs++;
			}

			/* --- If game was missing in dat1 --- */

			if (gamediff>0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games2[j].name, dat1.name);
				j++; total_diffs++;
			}
		}
	}

	/* --- If either dat uses alternative ROM names, use sort method 2 --- */

	if (!errflg && (dat1.flags & ALT_ROM_SUPPORT | dat2.flags & ALT_ROM_SUPPORT))
	{
		games1=dat1.games; games2=dat2.games;

		/* --- For all games --- */

		while (i<dat1.num_games || j<dat2.num_games)
		{
			/* --- If reached the end of either dat there is a game missing --- */

			if (i>=dat1.num_games) gamediff=1;
			if (j>=dat2.num_games) gamediff=-1;
			if (i<dat1.num_games && j<dat2.num_games)

			/* --- Else compare the game names --- */

				gamediff=strcmp(games1[i].name, games2[j].name);

			/* --- If the game names are the same then compare their details --- */

			if (gamediff==0)
			{
				/* --- Load the ROMs from dat1 into buffers  --- */
				/* --- One buffer for normal ROM details, one for alt names  --- */

				num_roms1=0;
				rom_ptr=games1[i].first_rom;
				alt_rom_ptr=games1[i].first_alt_rom;
				while (rom_ptr)
				{
					strcpy(roms1[num_roms1].name, rom_ptr->name);
					strcpy(roms1[num_roms1].size, rom_ptr->size);
					strcpy(roms1[num_roms1].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;

					/* --- If it has alt names, load them in buffer --- */

					if (dat1.flags & ALT_ROM_SUPPORT)
					{
						for (k=0; k<8; k++)
						{
							strcpy(alt_roms1[num_roms1].name[k], alt_rom_ptr->name[k]);
							strcpy(alt_roms1[num_roms1].size[k], alt_rom_ptr->size[k]);
						}
						alt_rom_ptr=alt_rom_ptr->next_alt_rom;
					}

					/* --- Else just store single name and clear alt names --- */

					else
					{
						strcpy(alt_roms1[num_roms1].name[0], roms1[num_roms1].name);
						strcpy(alt_roms1[num_roms1].size[0], roms1[num_roms1].size);
						for (k=1; k<8; k++)
						{
							alt_roms1[num_roms1].name[k][0]='\0';
							alt_roms1[num_roms1].size[k][0]='\0';
						}
					}

					num_roms1++;
				}

				/* --- Load the ROMs from dat2 into buffers  --- */
				/* --- One buffer for normal ROM details, one for alt names  --- */

				num_roms2=0;
				rom_ptr=games2[j].first_rom;
				alt_rom_ptr=games2[j].first_alt_rom;
				while (rom_ptr)
				{
					strcpy(roms2[num_roms2].name, rom_ptr->name);
					strcpy(roms2[num_roms2].size, rom_ptr->size);
					strcpy(roms2[num_roms2].crc, rom_ptr->crc);
					rom_ptr=rom_ptr->next_rom;

					/* --- If it has alt names, load them in buffer --- */

					if (dat2.flags & ALT_ROM_SUPPORT)
					{
						for (k=0; k<8; k++)
						{
							strcpy(alt_roms2[num_roms2].name[k], alt_rom_ptr->name[k]);
							strcpy(alt_roms2[num_roms2].size[k], alt_rom_ptr->size[k]);
						}
						alt_rom_ptr=alt_rom_ptr->next_alt_rom;
					}

					/* --- Else just store single name and clear alt names --- */

					else
					{
						strcpy(alt_roms2[num_roms2].name[0], roms2[num_roms2].name);
						strcpy(alt_roms2[num_roms2].size[0], roms2[num_roms2].size);
						for (k=1; k<8; k++)
						{
							alt_roms2[num_roms2].name[k][0]='\0';
							alt_roms2[num_roms2].size[k][0]='\0';
						}
					}

					num_roms2++;
				}

				/* --- For each ROM in buffer 1, check it is also in buffer 2 --- */

				for (k=0; k<num_roms1; k++)
				{
					found=0; sizediff=1;
					for (l=0; l<num_roms2; l++)
						for (m=0; alt_roms1[k].name[m][0]!='\0' && m<8; m++)
							for (n=0; alt_roms2[l].name[n][0]!='\0' && n<8; n++)
								if (!strcmp(alt_roms1[k].name[m], alt_roms2[l].name[n]))
								{
									found=1;
									if (alt_roms1[k].size[m][0]!='\0' && alt_roms2[l].size[n][0]!='\0')
										if (!strcmp(alt_roms1[k].size[m], alt_roms2[l].size[n]))
											sizediff=0;
										else
										{
											sprintf(size1, alt_roms1[k].size[m]);
											sprintf(size2, alt_roms2[l].size[n]);
										}
									else
										sizediff=0;

									if ((dat1.flags & dat2.flags & SIZE_SUPPORT) && roms1[k].size[0]=='\0')
										bad_size1++;

									if ((dat2.flags & dat2.flags & SIZE_SUPPORT) && roms2[l].size[0]=='\0')
										bad_size2++;
								}

					/* --- If not found, say so! --- */

					if (!found)
					{
						if (dat1.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games1[i].name, roms1[k].name, roms1[k].crc, dat2.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games1[i].name, roms1[k].name, dat2.name);
						total_diffs++;
					}

					/* --- If size differed, say so! --- */

					if (found && sizediff)
					{
						fprintf(out, "%s: %s has different sizes (%s - %s)\n", games1[i].name, roms1[k].name, size1, size2);
						total_diffs++;
					}
				}
				
				/* --- For each ROM in buffer 2, check it is also in buffer 1 --- */

				for (k=0; k<num_roms2; k++)
				{
					found=0;
					for (l=0; l<num_roms1; l++)
						for (m=0; alt_roms2[k].name[m][0]!='\0' && m<8; m++)
							for (n=0; alt_roms1[l].name[n][0]!='\0' && n<8; n++)
								if (!strcmp(alt_roms2[k].name[m], alt_roms1[l].name[n]))
									found=1;

					/* --- If not found, say so! --- */

					if (!found)
					{
						if (dat2.flags & CRC_SUPPORT)
							fprintf(out, "%s: %s [%s] missing in %s\n", games1[i].name, roms2[k].name, roms2[k].crc, dat1.name);
						else
							fprintf(out, "%s: %s missing in %s\n", games1[i].name, roms2[k].name, dat1.name);
						total_diffs++;
					}

					/* --- No need to report differences, they were reported earlier --- */
				}
				
				/* --- Onto the next game in each dat --- */

				i++; j++;
			}

			/* --- If game was missing in dat2 --- */

			if (gamediff<0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games1[i].name, dat2.name);
				i++; total_diffs++;
			}

			/* --- If game was missing in dat1 --- */

			if (gamediff>0)
			{
				fprintf(out, "%s: **game** missing in %s\n", games2[j].name, dat1.name);
				j++; total_diffs++;
			}
		}
	}

	/* --- Completion message --- */

	if (!errflg)
	{
		if (bad_romof1 || bad_romof2 || bad_size1 || bad_size2 || bad_merge1 || bad_merge2 ||
			dat1.num_samples || dat2.num_samples)
			fprintf(out, "\n");

		if (bad_romof1)	
		{
			fprintf(out, "Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof1, dat2.name);
			printf("Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof1, dat2.name);
		}

		if (bad_romof2)	
		{
			fprintf(out, "Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof2, dat1.name);
			printf("Warning: %d games in %s use differing cloneof/romof names.\n", bad_romof2, dat1.name);
		}

		if (bad_size1)	
		{
			fprintf(out, "Warning: %d ROMs in %s don't specify their size.\n", bad_size1, dat1.name);
			printf("Warning: %d ROMs in %s don't specify their size.\n", bad_size1, dat1.name);
		}

		if (bad_size2)	
		{
			fprintf(out, "Warning: %d ROMs in %s don't specify their size.\n", bad_size2, dat2.name);
			printf("Warning: %d ROMs in %s don't specify their size.\n", bad_size2, dat2.name);
		}

		if (bad_merge1)	
		{
			fprintf(out, "Warning: %d ROMs in %s use differing merge names.\n", bad_merge1, dat2.name);
			printf("Warning: %d ROMs in %s use differing merge names.\n", bad_merge1, dat2.name);
		}

		if (bad_merge2)	
		{
			fprintf(out, "Warning: %d ROMs in %s use differing merge names.\n", bad_merge2, dat1.name);
			printf("Warning: %d ROMs in %s use differing merge names.\n", bad_merge2, dat1.name);
		}

		if (dat1.num_samples || dat2.num_samples)	
		{
			fprintf(out, "Warning: Sample comparison not supported.\n");
			printf("Warning: Sample comparison not supported.\n");
		}

		if (bad_romof1 || bad_romof2 || bad_size1 || bad_size2 || bad_merge1 || bad_merge2 ||
			dat1.num_samples || dat2.num_samples)
		{
			fprintf(out, "\n");
			printf("\n");
		}

		if (total_diffs)
		{
			fprintf(out, "Total of %d differences found.\n\n", total_diffs);
			printf("Comparison complete. %d differences reported in %s\n", total_diffs, outfile);
		}
		else
		{
			fprintf(out, "No differences found.\n\n", total_diffs);
			printf("Comparison complete. No differences found.\n");
		}

		/* --- Close log file --- */

		fclose(out);
	}
}

/* --------------------------------------------------------------------------
 * Sorting functions for qsort - not very complicated!
 * -------------------------------------------------------------------------- */

int game_sort_function(const void *game1, const void *game2)
{
	return(strcmp(game1, game2));
}

int parent_sort_function(const void *game1, const void *game2)
{
	char *name1=((struct game *)(game1))->parent_name;
	char *name2=((struct game *)(game2))->parent_name;

	if (!*name1)
		name1=((struct game *)(game1))->name;

	if (!*name2)
		name2=((struct game *)(game2))->name;

	return(strcmp(name1, name2));
}

int rom_sort_function(const void *rom1, const void *rom2)
{
	return(strcmp(rom1, rom2));
}

/* --------------------------------------------------------------------------
 * Dat initialization - Loads the first ten lines of the dat into a buffer
 * and calls each of the identification functions. If the dat is successfully
 * identified, the structure is then ready for the load routines.
 * -------------------------------------------------------------------------- */

init_dat(struct dat *dat_info, char *fn)
{
	FILE *in;
	char lines[20][BUFFER_SIZE];
	int i=0, matches=0;

	/* --- Initialize dat structure --- */

	strcpy(dat_info->name, fn);
	dat_info->num_games=dat_info->num_parents=dat_info->num_clones=0;
	dat_info->flags=dat_info->num_roms=dat_info->num_resource_roms=0;
	dat_info->num_samples=0;
	dat_info->games=0; dat_info->roms=0; dat_info->samples=0;

	/* --- Open the dat file --- */

	if (!(in=fopen(dat_info->name, "r")))
	{
		printf("Error opening %s\n", dat_info->name);
		return(1);
	}

	/* --- Load the first ten lines --- */

	while (i<20 && fgets(lines[i], BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(lines[i]);
		i++;
	}

	/* --- Close the dat file --- */

	fclose(in);

	/* --- For all supported dat types, check if dat file matches --- */

	for (i=0;formats[i];i++)
	{
		if ((formats[i]->identify)((char *)lines))
		{
			sprintf(dat_info->descr, formats[i]->descr);
			dat_info->flags=formats[i]->flags;
			dat_info->load=formats[i]->load;
			dat_info->save=formats[i]->save;
			matches++;
		}
	}

	/* --- If no matches, dat is not of supported format --- */

	if (matches==0)
		printf("Can't identify type for '%s'\n", dat_info->name);

	/* --- If >1 matches, blank flags to prevent further processing --- */

	if (matches>1)
	{
		printf("Can't identify type for '%s'. Matches more than one type.\n", dat_info->name);
		dat_info->flags=0;
	}
}

/* --------------------------------------------------------------------------
 * Function to identify parents of games in memory.
 * This is to make life easier later (when writing out a RomCenter dat).
 * Also identifies Neo-Geo games where this can only be seen from the parent.
 * -------------------------------------------------------------------------- */

int identify_parents(struct dat *dat_info, char *game)
{
	struct game *games=dat_info->games;
	int num_games=dat_info->num_games;

	int num_parents=0;
	int unknown_parents=0;
	int i,j;
	int found;

	/* --- Unknown parents is allowable for formats such as System16 GCS --- */
	/* --- ...or there the user has specified an individual game --- */

	if ((dat_info->flags & ONE_GAME_ONLY) && games[0].parent_name[0]!='\0'||
		num_games==1 &&
		!strcmp(games[0].name, game) && strcmp(games[0].name, games[0].parent_name) && games[0].parent_name[0]!='\0')
	{
		dat_info->num_parents=0;
		printf("Warning: %s specifies an unknown parent (%s)\n\n",
			games[0].name, games[0].parent_name);
		return(0);
	}

	/* --- Otherwise, unknown parents is treated as a critical error --- */

	for (i=0;i<num_games; i++)
	{
		/* --- No point specifying this logic in every loader so it is here --- */

		if (!strcmp(games[i].name, "neogeo"))
			games[i].flags|= NEOGEO_GAME;

		/* --- If game specifies a parent --- */

		if (games[i].parent_name[0])
		{
			found=0;
 
			/* --- Search backward first (parent is likely to be before clone) --- */

			j=i-1;
			while (!found && j>=0)
			{
				if (!strcmp(games[j].name, games[i].parent_name))
				{
					strncpy(games[i].parent_title, games[j].title, TITLE_LENGTH);
					games[i].flags|=(games[j].flags & NEOGEO_GAME);

					/* --- If parent game not yet labelled as parent, do so --- */

					if (!(games[j].flags & PARENT_GAME))
					{
						games[j].flags|=PARENT_GAME;
						num_parents++;
					}
					found++;
				}

				j--;
			}

			/* --- Search forward if that didn't find it --- */

			j=i+1;
			while (!found && j<num_games)
			{
				if (!strcmp(games[j].name, games[i].parent_name))
				{
					strncpy(games[i].parent_title, games[j].title, TITLE_LENGTH);
					games[i].flags|=(games[j].flags & NEOGEO_GAME);

					/* --- If parent game not yet labelled as parent, do so --- */

					if (!(games[j].flags & PARENT_GAME))
					{
						games[j].flags|=PARENT_GAME;
						num_parents++;
					}
					found++;
				}

				j++;
			}

			if (!found)
			{
				/* --- IBM PC isn't declared in the MESS listinfo output--- */
 				if (!strcmp(games[i].parent_name, "pc"))
					sprintf(games[i].parent_title, "IBM PC");
				else
				{
					sprintf(games[i].parent_title, "Unknown");
					games[i].flags|=INVALID_PARENT;
				}
			}
		}
	}

	/* --- Update details in dat header for number of parents --- */

	dat_info->num_parents=num_parents;

	/* --- Return > 0 indicates a fatal error --- */

	return(unknown_parents);
}

/* --------------------------------------------------------------------------
 * Function to display summary information about the dat held in memory
 * -------------------------------------------------------------------------- */

display_dat_info(struct dat *dat_info)
{
	printf("Name:      %s\n", dat_info->name);
	printf("Type:      %s\n\n", dat_info->descr);

	printf("Games:     %d (%d parents, %d clones, %d other)\n", dat_info->num_games-dat_info->num_resources, dat_info->num_parents, dat_info->num_clones, dat_info->num_games-dat_info->num_parents-dat_info->num_clones-dat_info->num_resources);
	printf("Resources: %d\n", dat_info->num_resources);
	printf("ROMs:      %d (%d for resources)\n", dat_info->num_roms, dat_info->num_resource_roms);
	printf("Samples:   %d\n", dat_info->num_samples);
	printf("-------------------------------------------------------------------------------\n");
}

/* --------------------------------------------------------------------------
 * Function to release dat from memory
 * -------------------------------------------------------------------------- */

release_dat(struct dat *dat_info)
{
	if (dat_info->samples) cfree(dat_info->samples);
	if (dat_info->alt_roms) cfree(dat_info->alt_roms);
	if (dat_info->roms) cfree(dat_info->roms);
	if (dat_info->games) cfree(dat_info->games);

	return(0);
}

/* --------------------------------------------------------------------------
 * MAME Listinfo - Load and Save
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_mame_listinfo(char *lines)
{
	int i;

	/* --- If line exists with 'game (' or 'resource (', not 'emulator (' --- */

	for (i=0; i<10; i++)
	{
		if (!strncmp(lines+i*BUFFER_SIZE, "game (", 6))
			return(1);
		if (!strncmp(lines+i*BUFFER_SIZE, "resource (", 10))
			return(1);
		if (!strncmp(lines+i*BUFFER_SIZE, "emulator (", 10))
			return(0);
	}

	return(0);
}

/* --- Load --- */

int load_mame_listinfo(struct dat *dat_info, char *game, int clones)
{
	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (!strncmp(st_ptr, "resource", 8))
			HOLD_GAME_FLAG(RESOURCE)                 /* Neo-Geo ROMs */

		if (!strncmp(st_ptr, "machine", 7))
			HOLD_GAME_FLAG(MACHINE)                  /* Used by MESS */

		if (!strncmp(st_ptr, "name", 4))
		{
			GET_TOKEN(token, st_ptr)                 /* skip name token */
			GET_TOKEN(token, st_ptr)
			HOLD_GAME_NAME(token)                    /* Remember the game name */
		}

		if (!strncmp(st_ptr, "description", 11))
		{
			GET_TOKEN(token, st_ptr)                 /* skip description token */
			GET_TOKEN(token, st_ptr)
			HOLD_GAME_TITLE(token)                   /* Hold description */
		}

		if (!strncmp(st_ptr, "cloneof", 7))
		{
			GET_TOKEN(token, st_ptr)                 /* skip cloneof token */
			GET_TOKEN(token, st_ptr)
			HOLD_PARENT_NAME(token)                  /* ...and parent name */
		}
		

		if (!strncmp(st_ptr, "romof", 5))
		{
			GET_TOKEN(token, st_ptr)                 /* skip romof token */
			GET_TOKEN(token, st_ptr)
			HOLD_ROMOF(token)                        /* ...and romof */
		}

		if (!strncmp(st_ptr, "rom (", 5))
			STORE_ROM(st_ptr)                        /* Only stores if interested! */

		if (!strncmp(st_ptr, "sampleof", 8))
		{
			GET_TOKEN(token, st_ptr)                 /* skip sampleof token */
			GET_TOKEN(token, st_ptr)
			HOLD_SAMPLEOF(token)                     /* ...and sampleof */
		}

		if (!strncmp(st_ptr, "sample ", 7))
			STORE_SAMPLE(st_ptr+7)                   /* Only stores if interested! */

		if (st_ptr[0]==')')                        /* No more info for game */
			STORE_GAME                               /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --- Save --- */

int save_mame_listinfo(struct dat *dat_info, char *outfile, char *outmode)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i, j, bits;
	long size;
	int funny_size=0, bad_size=0, bad_parent=0, warning=1;
	struct rom *curr_rom;
	struct sample *curr_sample;

	/* --- Open file for writing --- */

	if (!(out=fopen(outfile, outmode)))
	{
		printf("Error opening %s\n", outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on --- */

	if (outmode[0]=='w')
		printf("Writing");
	else
		printf("Appending to");

	printf(" %s in MAME Listinfo format...\n", outfile);

	/* --- For each game... --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- It may be a game or resource (neogeo) --- */

		if (games[i].flags & RESOURCE)
			fprintf(out, "resource (\n");
		/*else if (games[i].flags & MACHINE) - Not ClrMame compatible
			fprintf(out, "machine (\n");*/
		else
			fprintf(out, "game (\n");

		/* --- Write its name, description, parent etc --- */

		if (strchr(games[i].name, ' '))
			fprintf(out, "\tname \"%s\"\n", games[i].name);
		else
			fprintf(out, "\tname %s\n", games[i].name);
		fprintf(out, "\tdescription \"%s\"\n", games[i].title);
		if (games[i].parent_name[0])
			if (strchr(games[i].parent_name, ' '))
				fprintf(out, "\tcloneof \"%s\"\n", games[i].parent_name);
			else
				fprintf(out, "\tcloneof %s\n", games[i].parent_name);
		if (games[i].romof[0])
			if (strchr(games[i].romof, ' '))
				fprintf(out, "\tromof \"%s\"\n", games[i].romof);
			else
				fprintf(out, "\tromof %s\n", games[i].romof);

		/* --- Write out its ROM details --- */

		if (curr_rom=games[i].first_rom)
		{
			while (curr_rom)
			{
				if (strchr(curr_rom->name, ' '))
					fprintf(out, "\trom ( name \"%s\" ", curr_rom->name);
				else
					fprintf(out, "\trom ( name %s ", curr_rom->name);

				/* --- Merged ROMs have an extra bit --- */

				if (curr_rom->merge[0])
					if (strchr(curr_rom->merge, ' '))
						fprintf(out, "merge \"%s\" ", curr_rom->merge);
					else
						fprintf(out, "merge %s ", curr_rom->merge);

				if (curr_rom->size[0])
				{
					size=atoi(curr_rom->size); bits=0;
					for (j=0; j < 32; j++)
					{
						if ((1<<j) & size)
							bits++;
					}
					if (bits!=1)
					{
						fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
						funny_size++;
					}
					fprintf(out, "size %s ", curr_rom->size);
				}
				else
				{
					fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
					bad_size++;
				}

				fprintf(out, "crc %s )\n", curr_rom->crc);

				curr_rom=curr_rom->next_rom;
			}
		}

		/* --- Write out its Sample details --- */

		if (games[i].sampleof[0])
			fprintf(out, "\tsampleof %s\n", games[i].sampleof);

		if (curr_sample=games[i].first_sample)
		{
			while (curr_sample)
			{
				fprintf(out, "\tsample %s\n",curr_sample->name);
				curr_sample=curr_sample->next_sample;
			}
		}
		
		/* --- End of details for the game --- */

		fprintf(out, ")\n\n");
	}

	/* --- Close file --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}
	}

	printf("\nAll done!\n");
}

/* --------------------------------------------------------------------------
 * MESS Listinfo - Load only (use MAME Listinfo for saving)
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_mess_listinfo(char *lines)
{
	int i;

	/* --- If first line begins with 'machine (' --- */

	for (i=0; i<10; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "machine (", 9))
			return(1);

	return(0);
}

/* --- Load (use MAME Listinfo loader) --- */

int load_mess_listinfo(struct dat *dat_info, char *game, int clones)
{
	return (load_mame_listinfo(dat_info, game, clones));
}

/* --- Save (use MAME Listinfo saver) --- */

int save_mess_listinfo(struct dat *dat_info, char *outfile, char *outmode)
{
	return(save_mame_listinfo(dat_info, outfile, outmode));
}

/* --------------------------------------------------------------------------
 * Shark Listinfo
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_shark_listinfo(char *lines)
{
	/* --- If first line begins with 'emulator ( "Shark' --- */

	if (!strncmp(lines, "emulator ( \"Shark", 16))
		return(1);

	return(0);
}

/* --- Load --- */

int load_shark_listinfo(struct dat *dat_info, char *game, int clones)
{
	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (!strncmp(st_ptr, "name", 4))
			HOLD_GAME_NAME(st_ptr+5)                 /* Remember the game name */

		if (!strncmp(st_ptr, "description", 11))
		{
			REMOVE_QUOTES(st_ptr+12)                 /* Remove quotes (if required) */
			HOLD_GAME_TITLE(st_ptr+12)               /* Hold description */
		}

		if (!strncmp(st_ptr, "rom (", 5))
		{
			LOWER(st_ptr)
			STORE_ROM(st_ptr)                        /* Only stores if interested! */
		}

		if (st_ptr[0]==')')                        /* No more info for this game */
			STORE_GAME  				                     /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * RAINE Gameinfo
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_raine_gameinfo(char *lines)
{
	/* --- If first line is 'emulator (' --- */
	/* --- and second contains 'name "Raine"' --- */

	if (!strcmp(lines, "emulator (") &&
			strstr(lines+BUFFER_SIZE, "name \"Raine\""))
		return(1);

	return(0);
}

/* --- Load --- */

int load_raine_gameinfo(struct dat *dat_info, char *game, int clones)
{
	int emulator=0;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		SKIP_WHITESPACE(st, st_ptr)                /* Some dats have indentation */

		if (!strncmp(st_ptr, "emulator (", 10))
			emulator=1;                              /* We aren't interested */

		if (!strncmp(st_ptr, "name", 4))
			HOLD_GAME_NAME(st_ptr+5)                 /* Remember the game name */

		if (!strncmp(st_ptr, "description", 11))
		{
			if (strstr(st_ptr, " japanese \""))
				*strstr(st_ptr, " japanese \"")='\0';  /* Skip Japenese version */
			REMOVE_QUOTES(st_ptr+12)                 /* Remove quotes (if required) */
			HOLD_GAME_TITLE(st_ptr+12)               /* Hold description */
		}

		if (!strncmp(st_ptr, "cloneof", 7))
			HOLD_PARENT_NAME(st_ptr+8)               /* ...and parent name */

		if (!strncmp(st_ptr, "romof", 5))
			HOLD_ROMOF(st_ptr+6)                     /* ...and romof */

		if (!strncmp(st_ptr, "rom (", 5))
			STORE_ROM(st_ptr)                        /* Only stores if interested! */

		if (st_ptr[0]==')')                        /* No more info for this game */
			if (!emulator)
				STORE_GAME			                       /* Only stores if interested! */
			else
				emulator=0;                            /* We are past the emulator */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Callus Gamelist
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_callus_gamelist(char *lines)
{
	/* --- If second line contains 'Callus' --- */

	if (strstr(lines+BUFFER_SIZE, "Callus"))
		return(1);

	return(0);
}

/* --- Load --- */

int load_callus_gamelist(struct dat *dat_info, char *game, int clones)
{
	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (st[0]==' ')                          /* Game title preceded by spaces */
			HOLD_GAME_TITLE(st)                    /* ...remember it! */

		if (!strncmp(st, "Name:", 5))
			HOLD_GAME_NAME(st+14)                  /* ...and the game name */

		if (!strncmp(st, "Directory:", 10))
			HOLD_PARENT_NAME(st+14)                /* ...and the parent name */

		if (st[0]=='	')                         /* ROM names are preceded by TAB */
		{
			LOWER(st)
			STORE_ROM(st+1)                        /* Only stores if interested! */
		}

		if (st[0]=='-')                          /* No more info for this game */
			STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * JFF Text
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_jff_txt(char *lines)
{
	int i;

	/* --- If any of the first ten lines contain 'JUST FOR FUN' --- */

	for (i=0; i<10; i++)
		if (strstr(lines+BUFFER_SIZE*i, "JUST FOR FUN"))
			return(1);

	return(0);
}

/* --- Load --- */

int load_jff_txt(struct dat *dat_info, char *game, int clones)
{
	int rom_flag=0;
	char rom_string[80];
	char rom_name[FILE_NAME_LENGTH+1];
	char rom_size[20];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strncmp(st, "###", 3))              /* Game title preceded by hashes */
		{
			st[strlen(st)-4]='\0';
			HOLD_GAME_TITLE(st+4)                  /* ...remember it! */
		}

		if (!strcmp(st+4, "directory name:"))
		{
			fgets(st, BUFFER_SIZE, in);
			fgets(st, BUFFER_SIZE, in);
			REMOVE_CR_LF(st)
			LOWER(st)
			HOLD_GAME_NAME(st)                     /* ...and the game name */
		}

		if (!strcmp(st+4, "names:"))
		{
			fgets(st, BUFFER_SIZE, in);
			REMOVE_CR_LF(st)
			while (st[0]=='\0')
			{
				fgets(st, BUFFER_SIZE, in);
				REMOVE_CR_LF(st)
			}
			rom_flag=1;                            /* next few lines are ROMs */
		}

		if (rom_flag && st[0]!='\0')             /* This is a ROM */
		{
			if (strchr(st, ';'))
				*strchr(st, ';')='\0';
			LOWER(st)
			EXTRACT_NAME(rom_name, st)
			EXTRACT_SIZE(rom_size, st+14)
			sprintf(rom_string, "rom ( name %s size %s )", rom_name, rom_size);
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}

		if (rom_flag && st[0]=='\0')             /* No more info for this game */
		{	
			STORE_GAME                             /* Only stores if interested! */
			rom_flag=0;
		}
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Sparcade Romlist - Not enabled as romlist.txt is not 100% correct.
 * It doesn't specify extensions correctly for some games, some game names are
 * incorrect and it misses some ROMs. These are all correct in the .RAT files.
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_sparcade_romlist(char *lines)
{
	int i;

	/* --- If any of the first ten lines contain 'Sparcade' --- */

	for (i=0; i<10; i++)
		if (strstr(lines+BUFFER_SIZE*i, "Sparcade"))
			return(1);

	return(0);
}

/* --- Load --- */

int load_sparcade_romlist(struct dat *dat_info, char *game, int clones)
{
	int past_example=0;
	char rom_string[120];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strcmp(st, "-----"))                 /* Past the example */
			past_example=1;

		if (past_example && strchr(st, '['))
		{
			if (strchr(st, ']'))
				*strrchr(st, ']')='\0';
			HOLD_GAME_NAME(strchr(st, '[')+1)      /* hold the game name */
			*strrchr(st, ' ')='\0';
			HOLD_GAME_TITLE(st);                   /* ...and the game title */
			fgets(st, BUFFER_SIZE, in);            /* skip the next line */
			fgets(st, BUFFER_SIZE, in);            /* get the next line */
			REMOVE_CR_LF(st)
		}

		if (past_example && !strncmp(st, "  ", 2))  /* This is a ROM */
		{
			LOWER(st)
			EXTRACT_NAMES(rom_string, st, "   ")   /* Ensure there is only one space between names */
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}

		if (past_example && st[0]=='\0')         /* No more info for this game */
			STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Sparcade RAT File
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_sparcade_rat(char *lines)
{
	int i;

	/* --- If any of the first ten lines begin 'path:' --- */

	for (i=0; i<10; i++)
		if (!strncmp(lines+BUFFER_SIZE*i, "path:", 5))
			return(1);

	return(0);
}

/* --- Load --- */

int load_sparcade_rat(struct dat *dat_info, char *game, int clones)
{
	char *name_ptr;
	char rom_string[120];
	char ext[10]="";

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		RTRIM(st)

		if (!strncmp(st, "path:", 5))
		{
	    STORE_GAME                             /* Only stores if interested! */
			ext[0]='\0';
			name_ptr=st+5;
			while (name_ptr[0]==' ' || name_ptr[0]=='	')
				name_ptr++;
			HOLD_GAME_NAME(name_ptr)               /* hold the game name */
		}

		if (!strncmp(st, "ext:", 4))
		{
			name_ptr=st+5;
			while (name_ptr[0]==' ' || name_ptr[0]=='	')
				name_ptr++;
			strncpy(ext, name_ptr, 3);             /* hold the game ext */
			ext[4]='\0';
		}

		if (st[0] && st[0]!=' ' && st[0]!='	' &&
        st[0]!='#' && (!strchr(st, ':')))    /* This is a ROM */
		{
			LOWER(st)
			EXTRACT_NAMES(rom_string, st, ext)     /* Ensure there is only one space between names */
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}
	}

	STORE_GAME                                 /* Only stores if interested! */

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * JAS-SYS1 Roms Text
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_jas_romlist(char *lines)
{
	/* --- Check text on the third line --- */

	if (!strncmp(lines+BUFFER_SIZE*2, "Directory", 9))
		return(1);

	return(0);
}

/* --- Load --- */

int load_jas_romlist(struct dat *dat_info, char *game, int clones)
{
	char *name_ptr;
	int rom_flag=0;
	char rom_string[80];
	char rom_name[FILE_NAME_LENGTH+1];
	char rom_size[20];

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)
		RTRIM(st)                                 /* There are some lines with just spaces */

		if (title_hold[0]=='\0' && st[0]!='\0')
			HOLD_GAME_TITLE(st)                     /* ...remember it! */

		if (!strncmp(st, "Directory", 9))
		{	
			LOWER(st)
			name_ptr=st+10;
			while (name_ptr[0]==' ' || name_ptr[0]=='	')
				name_ptr++;
			HOLD_GAME_NAME(name_ptr)               /* ...and the game name */
			fgets(st, BUFFER_SIZE, in);            /* skip the next line */
			fgets(st, BUFFER_SIZE, in);            /* skip the next line */
			fgets(st, BUFFER_SIZE, in);            /* get the next line */
			REMOVE_CR_LF(st)
			rom_flag=1;                            /* Okay, start looking for ROMs now! */
		}

		if (rom_flag && st[0]!='\0')
		{                                        /* This is a ROM */
			LOWER(st)
			EXTRACT_NAME(rom_name, st)
			EXTRACT_SIZE(rom_size, st+14)
			sprintf(rom_string, "rom ( name %s size %s )", rom_name, rom_size);
			STORE_ROM(rom_string)                  /* Only stores if interested! */
		}

		if (st[0]=='\0')                         /* No more info for this game */
		{	
				STORE_GAME                           /* Only stores if interested! */
				HOLD_GAME_TITLE("")
				rom_flag=0;
		}
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * Retrocade Romlist
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_retrocade_romlist(char *lines)
{
	/* --- If first line begins 'Game:' --- */

	if (!strncmp(lines, "Game:", 5))
		return(1);

	return(0);
}

/* --- Load --- */

int load_retrocade_romlist(struct dat *dat_info, char *game, int clones)
{
	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strncmp(st, "Game:", 5))
		{
			char *name=st+58;
			*strrchr(st, '.')='\0';        /* Remove '.zip )' from line */
			LOWER(st+58)
			HOLD_GAME_NAME(st+58)          /* Game name is now at end of line */
			*strrchr(st, '(')='\0';        /* Remove bracket infront of it */
			RTRIM(st)                      /* ...and the spaces before */
			HOLD_GAME_TITLE(st+6)          /* Game title is now at right of line */
		}

		if (!strncmp(st, "  ", 2))
		{
			*strchr(st, '(')='\0';         /* Remove status of rom */
			RTRIM(st)                      /* ...and spaces */
			LOWER(st)
			STORE_ROM_UNIQUE(st+8)         /* Only stores if interested! */
		}

		if (st[0]=='\0')                 /* Empty line signifies end of game */
			STORE_GAME                     /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * System16 GCS - Note that the shortname is incorrect in SHO_BOOT.GCS
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_system16_gcs(char *lines)
{
	int i;

	/* --- If any of the first ten lines begins with '.game' --- */

	for (i=0; i<10; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, ".game", 5))
			return(1);

	return(0);
}

/* --- Load --- */

int load_system16_gcs(struct dat *dat_info, char *game, int clones)
{
	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		if (!strncmp(st, ".game", 5))
			HOLD_GAME_TITLE(st+6)                  /* Remember the title */

		if (!strncmp(st, ".shortname", 10))
		{
			LOWER(st+11)
			HOLD_GAME_NAME(st+11)                  /* ...and the game name */
		}

		if (!strncmp(st, ".zip", 4))
		{
			*strrchr(st, '.')='\0';                /* Remove '.zip' */
			HOLD_PARENT_NAME(st+5)                 /* Remember the parent name */
		}

		if (!strncmp(st, ".romload", 8) ||
				!strncmp(st, ".scrload", 8) ||
				!strncmp(st, ".objload", 8) ||
				!strncmp(st, ".objload2", 9) ||
				!strncmp(st, ".z80load", 8))
			STORE_S16_ROMS(st, 1, 99)              /* Only stores if interested! */

		if (!strncmp(st, ".z80data", 8))
			STORE_S16_ROMS(st, 2, 99)              /* Only stores if interested! */

		if (!strncmp(st, ".loaddata", 9))
			STORE_S16_ROMS(st, 3, 99)              /* Only stores if interested! */

		if (!strncmp(st, ".grscrload", 10))
			STORE_S16_ROMS(st, 1, 1)               /* Only stores if interested! */

		if (!strncmp(st, ".end", 4))             /* No more info for this game */
			STORE_GAME                             /* Only stores if interested! */
	}

	DONE_LOAD
}

/* --------------------------------------------------------------------------
 * RomCenter Dat - Load and Save
 *
 * The loader doesn't use macros like everything else as RomCenter dats are so
 * different in nature to the others it has it's own algorithm!
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_romcenter_dat(char *lines)
{
	int i, found=0;

	/* --- If any of first ten lines begin with '[CREDITS]' --- */

	for (i=0; i<20; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "[CREDITS]", 9))
			found++;

	if (found)
		for (i=0; i<20; i++)
			if (!strncmp(lines+i*BUFFER_SIZE, "version=1", 9))
				return(1);
		
	return(0);
}

/* --- Load --- */

int load_romcenter_dat(struct dat *dat_info, char *game, int clones)
{
	struct rom *next_rom;

	char parent_name[GAME_NAME_LENGTH+1];
	char parent_title[TITLE_LENGTH+1];
	char game_name[GAME_NAME_LENGTH+1];
	char game_title[TITLE_LENGTH+1];
	char romof_name[GAME_NAME_LENGTH+1];
	char rom_name[FILE_NAME_LENGTH+1];
	long flags;
	char merge_name[FILE_NAME_LENGTH+1];
	char rom_size[SIZE_LENGTH+1];
	char rom_crc[CRC_LENGTH+1];

	char *ptr;

	int curr_game=0;
	int found_game, i;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		/* --- We're only interested in ROM lines --- */

		if (st[0]=='')
		{
			/* --- Extract parent name and game name --- */

			strncpy(parent_name, strtok(st,""), GAME_NAME_LENGTH);
			parent_name[GAME_NAME_LENGTH]='\0';
			strtok(NULL,"");
			strncpy(game_name, strtok(NULL,""), GAME_NAME_LENGTH);
			game_name[GAME_NAME_LENGTH]='\0';

			/* --- Are we interested in this line? --- */
			/* --- Dependant on -g and -c options --- */

			if (!game||
					game && !strcmp(game, game_name) ||
					game && clones && !strcmp(game, parent_name))
			{
				/* --- Hard coded logic to say neogeo is a resource --- */

				if (!strcmp(game_name, "neogeo"))
					flags=RESOURCE;
				else
					flags=0;

				/* --- Extract other details about game and ROM --- */

				strncpy(game_title, strtok(NULL,""), TITLE_LENGTH);
				game_title[TITLE_LENGTH]='\0';
				strncpy(rom_name, strtok(NULL,""), FILE_NAME_LENGTH);
				rom_name[FILE_NAME_LENGTH]='\0';
				strcpy(rom_crc, strtok(NULL,""));
				if (strcmp(strtok(NULL,""), "N"))
					flags|=NEOGEO_GAME;

				/* --- Best we can do for merge name --- */

				if (!strcmp(strtok(NULL,""), "N"))
					sprintf(merge_name, rom_name);
				else
					merge_name[0]='\0';

				/* --- Size may or may not be there --- */

				if (ptr=strtok(NULL,""))
					strcpy(rom_size, ptr);
				else
					rom_size[0]='\0';

				/* --- Store the ROM details --- */

				if (strcmp(rom_crc, "        "))
				{
					strcpy(roms[num_roms].name, rom_name);
					strcpy(roms[num_roms].merge, merge_name);
					strcpy(roms[num_roms].size, rom_size);
					strcpy(roms[num_roms].crc, rom_crc);
				}

				/* --- See if the game details are already stored --- */

				found_game=0;
				for (curr_game=num_games-1; !found_game && curr_game>=0; curr_game--)
				{
					if (!strcmp(game_name, games[curr_game].name))
						found_game=1;
				}

				/* --- If not... --- */

				if (!found_game)
				{
					/* --- Store game details --- */

					strcpy(games[num_games].name, game_name);
					strcpy(games[num_games].title, game_title);
					if (strcmp(game_name, parent_name))
					{
						strcpy(games[num_games].parent_name, parent_name);
						num_clones++;
					}

					/* --- If game has parent, assume 'romof' is the same --- */

					if (strcmp(parent_name, game_name))
						strcpy(games[num_games].romof, parent_name);
					else
						games[num_games].romof[0]='\0';

					/* --- Store the flags --- */

					games[num_games].flags=flags;

					/* --- If game is for Neo-Geo and romof not specified... --- */
					if ((games[num_games].flags & NEOGEO_GAME) &&
							strcmp(games[num_games].name, "neogeo") &&
							!games[num_games].romof[0])
						sprintf(games[num_games].romof, "neogeo");

					/* --- Let the game know where the first ROM is stored --- */

					if (strcmp(rom_crc, "        "))
						games[num_games].first_rom=&roms[num_roms];

					/* --- Increment game and resource counters --- */

					num_games++;

					if (flags & RESOURCE)
						num_resources++;
				}
				else
				{
					/* --- We found the game but have gone one past it --- */

					curr_game++;

					/* --- If it doesn't have any ROMs (actually impossible) --- */

					if (strcmp(rom_crc, "        "))
					{
						if (!games[curr_game].first_rom)
							games[curr_game].first_rom=&roms[num_roms];
						else
						{
							/* --- Walk through list until last ROM reached --- */

							next_rom=games[curr_game].first_rom;

							while (next_rom->next_rom)
								next_rom=next_rom->next_rom;

							/* --- Tell it where the last one is --- */
	
							next_rom->next_rom=&roms[num_roms];
						}
					}
				}

				/* --- One more ROM (and it may have been for a resource) --- */

				if (strcmp(rom_crc, "        "))
				{
					num_roms++;

					if (flags & RESOURCE)
						num_resource_roms++;
				}
			}
		}
	}

	DONE_LOAD
}

/* --- Save --- */

int save_romcenter_dat(struct dat *dat_info, char *outfile, char *outmode)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i, j, bits;
	long size;
	int bad_parent=0, funny_size=0, bad_size=0, bad_merge=0, bad_romof=0, warning=1;
	struct rom *curr_rom;

	/* --- Open file for writing --- */

	if (!(out=fopen(outfile, outmode)))
	{
		printf("Error opening %s\n", outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on and write header if required --- */

	if (outmode[0]=='w')
	{
		printf("Writing");
		fprintf(out, "[CREDITS]\n");
		fprintf(out, "Author=\n");
		fprintf(out, "Version=\n");
		fprintf(out, "Comment=Generated by DatUtil v1.3\n");
		fprintf(out, "[DAT]\n");
		fprintf(out, "version=1.90\n");
		fprintf(out, "[EMULATOR]\n");
		fprintf(out, "refname=\n");
		fprintf(out, "version=\n");
		fprintf(out, "Exe=\n");
		fprintf(out, "runcmd=\n");
		fprintf(out, "romspaths=\n");
		fprintf(out, "[GAMES]\n");
	}
	else
		printf("Appending to");

	printf(" %s in RomCenter format...\n", outfile);

	/* --- For each game --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- If 'romof' differs from parent, we can't represent this --- */

		if (strcmp(games[i].parent_name, games[i].romof) &&
				strcmp(games[i].romof, "neogeo"))
		{
			fprintf(log, "%s: specifies different cloneof and romof ('%s' and '%s')\n",
				games[i].name, games[i].parent_name, games[i].romof);
			bad_romof++;
		}

		/* --- If the game doesn't have a parent, then use the game itself --- */

		if (!games[i].parent_name[0])
		{
			strcpy(games[i].parent_name, games[i].name);
			strcpy(games[i].parent_title, games[i].title);
		}

		/* --- Write out the ROM details --- */

		if (curr_rom=games[i].first_rom)
		{
			while (curr_rom)
			{
				/* --- Don't keep Neo-Geo ROMs (only in the neogeo resource) --- */

				if ((games[i].flags & RESOURCE) ||
					strcmp(curr_rom->crc, "9036d879") &&
					strcmp(curr_rom->crc, "97cf998b") &&
					strcmp(curr_rom->crc, "354029fc"))
				{
					/* --- General details --- */

					fprintf(out, "%s%s%s%s%s%s", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title, curr_rom->name, curr_rom->crc);

					/* --- Neo-Geo flag and merge flag --- */

					if (games[i].flags & NEOGEO_GAME)
						fprintf(out, "O");
					else
						fprintf(out, "N");

					if (!curr_rom->merge[0])
						fprintf(out, "O");
					else
					{	
						/* --- If merge name differs, we can't represent this --- */

						if (strcmp(curr_rom->merge, curr_rom->name))
						{
							fprintf(log, "%s: %s has a different merge name ('%s')\n", games[i].name, curr_rom->name, curr_rom->merge);
							bad_merge++;
						}

						fprintf(out, "N");
					}

					/* --- The ROM size --- */

					if (curr_rom->size[0])
					{
						size=atoi(curr_rom->size); bits=0;
						for (j=0; j < 32; j++)
						{
							if ((1<<j) & size)
								bits++;
						}
						if (bits!=1)
						{
							fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
							funny_size++;
						}
						fprintf(out, "%s", curr_rom->size);
					}
					else
					{
						fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
						bad_size++;
					}

					fprintf(out, "\n");
				}

				curr_rom=curr_rom->next_rom;
			}
		}
		else
		{
			fprintf(out, "%s%s%s%sNO_ROMS        NO\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title);
		}
	}

	/* --- Close output files --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size || bad_merge || bad_romof || dat_info->num_samples)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}

		if (bad_merge)	
		{
			printf("%d) %d ROMs specify merge name different to their own.\n   This information cannot be represented in RomCenter datafiles.\n", warning, bad_merge);
			warning++;
		}

		if (bad_romof)	
		{
			printf("%d) %d games specify romof different to their parent.\n   This information cannot be represented in RomCenter datafiles.\n", warning, bad_romof);
			warning++;
		}

		if (dat_info->num_samples)	
		{
			printf("%d) %d samples were listed in the original file.\n   This information cannot be represented in RomCenter datafiles.\n", warning, dat_info->num_samples);
			warning++;
		}
	}

	printf("\nAll done!\n");
}


/* --------------------------------------------------------------------------
 * RomCenter 2 Dat - Load and Save
 *
 * The loader doesn't use macros like everything else as RomCenter dats are so
 * different in nature to the others it has it's own algorithm!
 * -------------------------------------------------------------------------- */

/* --- Identification --- */

int identify_romcenter2_dat(char *lines)
{
	int i, found=0;

	/* --- If any of first ten lines begin with '[CREDITS]' --- */

	for (i=0; i<20; i++)
		if (!strncmp(lines+i*BUFFER_SIZE, "[CREDITS]", 9))
			found++;

	if (found)
		for (i=0; i<20; i++)
			if (!strncmp(lines+i*BUFFER_SIZE, "version=2", 9))
				return(1);
		
	return(0);
}

/* --- Load --- */

int load_romcenter2_dat(struct dat *dat_info, char *game, int clones)
{
	struct rom *next_rom;

	char parent_name[GAME_NAME_LENGTH+1];
	char parent_title[TITLE_LENGTH+1];
	char game_name[GAME_NAME_LENGTH+1];
	char game_title[TITLE_LENGTH+1];
	char romof_name[GAME_NAME_LENGTH+1];
	char rom_name[FILE_NAME_LENGTH+1];
	long flags;
	char merge_name[FILE_NAME_LENGTH+1];
	char rom_size[SIZE_LENGTH+1];
	char rom_crc[CRC_LENGTH+1];

	char *ptr;

	int curr_game=0;
	int found_game, i;

	INIT_LOAD

	while (fgets(st, BUFFER_SIZE, in))
	{
		REMOVE_CR_LF(st)

		/* --- We're only interested in ROM lines --- */

		if (st[0]=='')
		{
			/* --- Extract parent name and game name --- */

			strncpy(parent_name, strtok(st,""), GAME_NAME_LENGTH);
			parent_name[GAME_NAME_LENGTH]='\0';
			strtok(NULL,"");
			strncpy(game_name, strtok(NULL,""), GAME_NAME_LENGTH);
			game_name[GAME_NAME_LENGTH]='\0';

			/* --- Are we interested in this line? --- */
			/* --- Dependant on -g and -c options --- */

			if (!game||
					game && !strcmp(game, game_name) ||
					game && clones && !strcmp(game, parent_name))
			{
				/* --- Hard coded logic to say neogeo is a resource --- */

				if (!strcmp(game_name, "neogeo"))
					flags=RESOURCE;
				else
					flags=0;

				/* --- Extract other details about game and ROM --- */

				strncpy(game_title, strtok(NULL,""), TITLE_LENGTH);
				game_title[TITLE_LENGTH]='\0';
				strncpy(rom_name, strtok(NULL,""), FILE_NAME_LENGTH);
				rom_name[FILE_NAME_LENGTH]='\0';
				strcpy(rom_crc, strtok(NULL,""));
				strncpy(rom_size, strtok(NULL,""), SIZE_LENGTH);
				rom_size[SIZE_LENGTH]='\0';
				ptr=strtok(NULL,"");
				if (ptr)
				{
					strncpy(romof_name, ptr, GAME_NAME_LENGTH);
					romof_name[GAME_NAME_LENGTH]='\0';
				}
				else
					*romof_name='\0';
				ptr=strtok(NULL,"");
				if (ptr)
				{
					strncpy(merge_name, ptr, FILE_NAME_LENGTH);
					merge_name[FILE_NAME_LENGTH]='\0';
				}
				else
					*merge_name='\0';

				/* --- Store the ROM details --- */

				if (strcmp(rom_crc, "        "))
				{
					strcpy(roms[num_roms].name, rom_name);
					strcpy(roms[num_roms].merge, merge_name);
					strcpy(roms[num_roms].size, rom_size);
					strcpy(roms[num_roms].crc, rom_crc);
				}

				/* --- See if the game details are already stored --- */

				found_game=0;
				for (curr_game=num_games-1; !found_game && curr_game>=0; curr_game--)
				{
					if (!strcmp(game_name, games[curr_game].name))
						found_game=1;
				}

				/* --- If not... --- */

				if (!found_game)
				{
					/* --- Store game details --- */

					strcpy(games[num_games].name, game_name);
					strcpy(games[num_games].title, game_title);
					if (strcmp(game_name, parent_name))
					{
						strcpy(games[num_games].parent_name, parent_name);
						num_clones++;
					}

					/* --- If game has parent, assume 'romof' is the same --- */

					strcpy(games[num_games].romof, romof_name);

					/* --- Store the flags --- */

					games[num_games].flags=flags;

					/* --- Let the game know where the first ROM is stored --- */

					if (strcmp(rom_crc, "        "))
						games[num_games].first_rom=&roms[num_roms];

					/* --- Increment game and resource counters --- */

					num_games++;

					if (flags & RESOURCE)
						num_resources++;
				}
				else
				{
					/* --- We found the game but have gone one past it --- */

					curr_game++;

					/* --- If it doesn't have any ROMs (actually impossible) --- */

					if (strcmp(rom_crc, "        "))
					{
						if (!games[curr_game].first_rom)
							games[curr_game].first_rom=&roms[num_roms];
						else
						{
							/* --- Walk through list until last ROM reached --- */

							next_rom=games[curr_game].first_rom;

							while (next_rom->next_rom)
								next_rom=next_rom->next_rom;

							/* --- Tell it where the last one is --- */
	
							next_rom->next_rom=&roms[num_roms];
						}
					}
				}

				/* --- One more ROM (and it may have been for a resource) --- */

				if (strcmp(rom_crc, "        "))
				{
					num_roms++;

					if (flags & RESOURCE)
						num_resource_roms++;
				}
			}
		}
	}

	DONE_LOAD
}

/* --- Save --- */

int save_romcenter2_dat(struct dat *dat_info, char *outfile, char *outmode)
{
	struct game *games=dat_info->games;
	FILE *out, *log;
	int i, j, bits;
	long size;
	int bad_parent=0, funny_size=0, bad_size=0, warning=1;
	struct rom *curr_rom;

	/* --- Open file for writing --- */

	if (!(out=fopen(outfile, outmode)))
	{
		printf("Error opening %s\n", outfile);
		return(1);
	}

	if (!(log=fopen("datutil.log", outmode)))
	{
		fclose(out);
		printf("Error opening %s\n", "datutil.log");
		return(1);
	}

	/* --- Tell user what is going on and write header if required --- */

	if (outmode[0]=='w')
	{
		printf("Writing");
		fprintf(out, "[CREDITS]\n");
		fprintf(out, "Author=\n");
		fprintf(out, "Version=\n");
		fprintf(out, "Comment=Generated by DatUtil v1.3\n");
		fprintf(out, "[DAT]\n");
		fprintf(out, "version=2.00\n");
		fprintf(out, "[EMULATOR]\n");
		fprintf(out, "refname=\n");
		fprintf(out, "version=\n");
		fprintf(out, "[GAMES]\n");
	}
	else
		printf("Appending to");

	printf(" %s in RomCenter 2 format...\n", outfile);

	/* --- For each game --- */

	for (i=0; i<dat_info->num_games; i++)
	{
		/* --- Check for invalid parent --- */
		if (games[i].flags & INVALID_PARENT)
		{
			fprintf(log, "%s: specifies an invalid parent ('%s')\n", games[i].name, games[i].parent_name);
			bad_parent++;
		}

		/* --- If the game doesn't have a parent, then use the game itself --- */

		if (!games[i].parent_name[0])
		{
			strcpy(games[i].parent_name, games[i].name);
			strcpy(games[i].parent_title, games[i].title);
		}

		/* --- Write out the ROM details --- */

		if (curr_rom=games[i].first_rom)
		{
			while (curr_rom)
			{
				/* --- General details --- */

				fprintf(out, "%s%s%s%s%s%s%s%s%s\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title, curr_rom->name, curr_rom->crc, curr_rom->size, games[i].romof, curr_rom->merge);

				/* --- The ROM size --- */

				if (curr_rom->size[0])
				{
					size=atoi(curr_rom->size); bits=0;
					for (j=0; j < 32; j++)
					{
						if ((1<<j) & size)
							bits++;
					}
					if (bits!=1)
					{
						fprintf(log, "%s: %s has a funny size ('%s')\n", games[i].name, curr_rom->name, curr_rom->size);
						funny_size++;
					}
				}
				else
				{
					fprintf(log, "%s: %s has no size\n", games[i].name, curr_rom->name);
					bad_size++;
				}

				curr_rom=curr_rom->next_rom;
			}
		}
		else
		{
			fprintf(out, "%s%s%s%sNO_ROMS        0\n", games[i].parent_name, games[i].parent_title, games[i].name, games[i].title);
		}
	}

	/* --- Close output files --- */

	fclose(log);
	fclose(out);

	/* --- Display summary to user --- */

	if (bad_parent || funny_size || bad_size || dat_info->num_samples)
	{
		printf("\nWarnings (see datutil.log):\n\n");

		if (bad_parent)	
		{
			printf("%d) %d games have an invalid parent.\n", warning, bad_parent);
			warning++;
		}

		if (funny_size)	
		{
			printf("%d) %d ROMs have a funny size.\n", warning, funny_size);
			warning++;
		}

		if (bad_size)	
		{
			printf("%d) %d ROMs don't specify their size.\n", warning, bad_size);
			warning++;
		}

		if (dat_info->num_samples)	
		{
			printf("%d) %d samples were listed in the original file.\n   This information cannot be represented in RomCenter 2 datafiles.\n", warning, dat_info->num_samples);
			warning++;
		}
	}

	printf("\nAll done!\n");
}
