/* --------------------------------------------------------------------------
 * DatUtil v1.3 (19 May 2000)
 * Written by Logiqx (logiqx@emuhq.com)
 *
 * A simple little utility for converting, cleansing and comparing ROM manager
 * dats (e.g. ClrMamePro, RomCenter). Also supports ROM lists generated by
 * emulators such as Retrocade, Callus and System 16 so that dats can be checked
 * for missing/spurious ROMs.
 *
 * Additional formats can easily be added if you have know a little C.
 * -------------------------------------------------------------------------- */

/* --- Size declarations --- */

#define MAX_GAMES 5000
#define MAX_ROMS 65536
#define MAX_SAMPLES 4096

#ifdef LONG_NAMES
#define GAME_NAME_LENGTH 80
#define FILE_NAME_LENGTH 80
#else
#define GAME_NAME_LENGTH 20
#define FILE_NAME_LENGTH 12
#endif

#define LONG_NAME_LENGTH 128
#define TITLE_LENGTH 100
#define CRC_LENGTH 8
#define SIZE_LENGTH 8

#define BUFFER_SIZE 1024

/* --- Game flags --- */

#define	NEOGEO_GAME			0x0001  /* Neo-Geo game flag */
#define	PARENT_GAME			0x0002  /* Parent game flag */
#define	RESOURCE				0x0004  /* Resource flag (used by MAME) */
#define	MACHINE					0x0008  /* Machine flag (used by MESS) */
#define	INVALID_PARENT	0x0010  /* Specifies an invalid parent */

/* --- Information that dat types can include --- */

#define	LOAD_SUPPORT		0x0001	/* DatUtil load support */
#define	SAVE_SUPPORT		0x0002	/* DatUtil save support */
#define	SIZE_SUPPORT		0x0004	/* ROM can specify size */
#define	CRC_SUPPORT			0x0008	/* ROM can specify CRC */
#define	PARENT_SUPPORT	0x0010	/* Game can specify parent of game */
#define	ROMOF_SUPPORT		0x0020	/* Game can specify where to find merge roms */
#define	MERGE_SUPPORT		0x0040	/* ROM can use different name from parent set */
#define	ONE_GAME_ONLY		0x0080	/* Dat can only support a single game */
#define	ALT_ROM_SUPPORT	0x0100	/* ROM can have one of several names */
#define	SAMPLE_SUPPORT	0x0200	/* Dat can include sample information  */

#define FULL_SUPPORT			LOAD_SUPPORT|SAVE_SUPPORT|SIZE_SUPPORT|CRC_SUPPORT| \
													PARENT_SUPPORT|ROMOF_SUPPORT|MERGE_SUPPORT| \
													SAMPLE_SUPPORT
#define ROMCENT_SUPPORT		LOAD_SUPPORT|SAVE_SUPPORT|SIZE_SUPPORT|CRC_SUPPORT| \
													PARENT_SUPPORT
#define ROMCENT2_SUPPORT	ROMCENT_SUPPORT|ROMOF_SUPPORT|MERGE_SUPPORT


/* --- The structures used to hold a dat in memory --- */

struct sample
{
	char name[FILE_NAME_LENGTH+1];
	struct sample *next_sample;
};

struct rom
{
	char name[FILE_NAME_LENGTH+1];
	char merge[FILE_NAME_LENGTH+1];
	char size[SIZE_LENGTH+1];
	char crc[CRC_LENGTH+1];
	struct rom *next_rom;
};

struct alt_rom
{
	char name[8][FILE_NAME_LENGTH+1];
	char size[8][SIZE_LENGTH+1];
	struct alt_rom *next_alt_rom;
};

struct game
{
	char name[GAME_NAME_LENGTH+1];
	char title[TITLE_LENGTH+1];
	char parent_name[GAME_NAME_LENGTH+1];		/* Helps saving RomCenter format */
	char parent_title[TITLE_LENGTH+1];
	char romof[GAME_NAME_LENGTH+1];
	char sampleof[GAME_NAME_LENGTH+1];
	long flags;
	struct rom *first_rom;
	struct alt_rom *first_alt_rom;
	struct sample *first_sample;
};

struct dat
{
	char name[LONG_NAME_LENGTH+1];					/* Filename of dat */
	char descr[40];
	long flags;
	int (*load)(struct dat *, char *, int);
	int (*save)(struct dat *, char *, char *);
	int num_games, num_parents, num_clones, num_resources;
	int num_samples;
	long num_roms, num_resource_roms;
	struct game *games;
	struct rom *roms;
	struct alt_rom *alt_roms;
	struct sample *samples;
};


/* --- The structure used to hold information of a dat format --- */

struct DatFormat
{
	char descr[40];
	char option[20];
	long flags;
	int (*identify)(char *);
	int (*load)(struct dat *, char *, int);
	int (*save)(struct dat *, char *, char *);
};

#define LOAD_SAVE_FORMAT(NAME, DESCR, OPTION, FLAGS) \
int identify_##NAME(char *); \
int load_##NAME(struct dat *, char *, int); \
int save_##NAME(struct dat *, char *, char *); \
struct DatFormat driver_##NAME = \
{ \
	DESCR, \
	OPTION, \
	LOAD_SUPPORT|SAVE_SUPPORT|FLAGS, \
	identify_##NAME, \
	load_##NAME, \
	save_##NAME \
};

#define LOAD_ONLY_FORMAT(NAME, DESCR, FLAGS) \
int identify_##NAME(char *); \
int load_##NAME(struct dat *, char *, int); \
struct DatFormat driver_##NAME = \
{ \
	DESCR, \
	"", \
	LOAD_SUPPORT|FLAGS, \
	identify_##NAME, \
	load_##NAME, \
	NULL \
};

#define FORMAT(NAME) &driver_##NAME,


/* --- Function prototypes for the sorting functions --- */

int game_sort_function(const void *, const void *);
int parent_sort_function(const void *, const void *);
int rom_sort_function(const void *, const void *);


/* --- Macro to load a dat into memory --- */

#define LOAD_DAT(DAT) \
{ \
	if (game[0]) \
		errflg+=(DAT.load)(&DAT, game, clones); \
	else \
		errflg+=(DAT.load)(&DAT, NULL, clones); \
}


/* --------------------------------------------------------------------------
 * Macros for string handling
 * -------------------------------------------------------------------------- */

#define EXTRACT_NAME(ROM_NAME, ST) \
{ \
	int i, j=0; \
\
	for (i=0; i<8; i++) \
		if (ST[i]!=' ') \
			ROM_NAME[j++]=ST[i]; \
\
	if (ST[9]!=' ' && ST[9]!='	') \
		ROM_NAME[j++]='.'; \
\
	for (i=9; i<12; i++) \
		if (ST[i]!=' ' && ST[i]!='	') \
			ROM_NAME[j++]=ST[i]; \
\
	ROM_NAME[j++]='\0'; \
}

#define EXTRACT_SIZE(ROM_SIZE, ST) \
{ \
	int i=0, j=0; \
\
	strncpy(ROM_SIZE, ST, 20); \
	while (ROM_SIZE[i]) \
	{ \
		if (ROM_SIZE[i]!=',' && ROM_SIZE[i]!=' ' && ROM_SIZE[i]!='	') \
			ROM_SIZE[j++]=ROM_SIZE[i++]; \
		else \
			i++; \
	} \
	ROM_SIZE[j]='\0'; \
}

/* --- For Sparcade --- */

#define EXTRACT_NAMES(ROM_STRING, ST, EXT) \
{ \
	int i=0, j=0, name_flag=0, ext_flag=0; \
\
	sprintf(ROM_STRING, "rom ( "); \
	j+=6; \
\
	while (ST[i]!=' ' && ST[i]!='	') \
		i++; \
\
	for (; i<strlen(ST); i++) \
	{ \
		if (ST[i]!=' ' && ST[i]!='	') \
		{ \
			if (!name_flag) \
			{ \
				strncpy(ROM_STRING+j, "name ", 5); \
				j+=5; \
			} \
			ROM_STRING[j++]=ST[i]; \
			if (ST[i]=='.') \
				ext_flag++; \
			name_flag++; \
		} \
		else if (name_flag) \
		{ \
			if (!ext_flag && EXT[0]!='\0') \
			{ \
				sprintf(ROM_STRING+j, ".%s", EXT); \
				j+=4; \
			} \
			ROM_STRING[j++]=' '; \
			name_flag=0; \
			ext_flag=0; \
		} \
	} \
\
	strcpy(ROM_STRING+j, " )"); \
}

#define REMOVE_QUOTES(ST) \
{ \
	if (*(ST)=='"') \
	{ \
		strcpy(ST, ST+1); \
		if (strrchr(ST, '"')) \
			*strrchr(ST, '"')='\0'; \
	} \
}

#define RTRIM(ST) \
{ \
	while (ST[strlen(ST)-1]==' ' || ST[strlen(ST)-1]=='	') \
		ST[strlen(ST)-1]='\0'; \
}

#define UPPER(ST) \
{ \
	int i; \
	for (i=0; i<strlen(ST); i++) \
		ST[i]=toupper(ST[i]); \
}

#define LOWER(ST) \
{ \
	char *ptr=ST; \
	int i; \
	for (i=0; i<strlen(ptr); i++) \
		ptr[i]=tolower(ptr[i]); \
}

#define REMOVE_CR_LF(ST) \
{ \
	while (ST[strlen(ST)-1]==10 || ST[strlen(ST)-1]==13) \
		ST[strlen(ST)-1]='\0'; \
}

#define SKIP_WHITESPACE(ST, ST_PTR) \
{ \
	ST_PTR=ST; \
	while (ST_PTR[0]=='	' || ST_PTR[0]==' ') ST_PTR++; \
}


/* --------------------------------------------------------------------------
 * INIT_LOAD macro
 *
 * 1) Standard variables used during loading
 * 2) Allocates memory for dat (not intelligent, just a large amount!)
 * 3) Opens dat file for reading
 * -------------------------------------------------------------------------- */

#define INIT_LOAD \
FILE *in; \
char st[BUFFER_SIZE], tmp[BUFFER_SIZE], token[BUFFER_SIZE]; \
char *st_ptr; \
char title_hold[TITLE_LENGTH+1]=""; \
char game_hold[GAME_NAME_LENGTH+1]=""; \
char parent_hold[GAME_NAME_LENGTH+1]=""; \
char romof_hold[GAME_NAME_LENGTH+1]=""; \
char sampleof_hold[GAME_NAME_LENGTH+1]=""; \
int flags_hold=0; \
struct game *games; struct rom *roms; struct sample *samples;\
struct alt_rom *alt_roms; \
struct rom *rom_ptr; \
int num_games=0, num_clones=0, num_resources=0, num_samples=0; \
long num_roms=0, num_resource_roms=0; \
int processing_game=0; \
int alt_name; \
{ \
\
	games=dat_info->games=calloc(MAX_GAMES, sizeof(struct game)); \
	roms=dat_info->roms=calloc(MAX_ROMS, sizeof(struct rom)); \
	alt_roms=dat_info->alt_roms=calloc(1024, sizeof(struct alt_rom)); \
	samples=dat_info->samples=calloc(MAX_SAMPLES, sizeof(struct sample)); \
\
	/*printf("Allocated %d bytes\n", MAX_GAMES*sizeof(struct game) + MAX_ROMS*sizeof(struct rom) + 1024*sizeof(struct alt_rom) + MAX_SAMPLES*sizeof(struct sample)); */ \
\
	if (!(games && roms && alt_roms && samples)) \
	{ \
		printf("Failed to allocate memory for dat\n"); \
		return(1); \
	} \
\
	if (!(in=fopen(dat_info->name, "r"))) \
	{ \
		printf("Error opening %s\n", dat_info->name); \
		return(1); \
	} \
}

/* --------------------------------------------------------------------------
 * STORE_GAME macro
 *
 * If current game/resource is being processed, it is stored in memory and
 * the game/resource counters incremented. Also clears out hold variables.
 * -------------------------------------------------------------------------- */

#define STORE_GAME \
{ \
	if (processing_game) \
	{ \
		strcpy(games[num_games].name, game_hold); \
		strcpy(games[num_games].title, title_hold); \
		if (parent_hold[0]!='\0' && strcmp(game_hold, parent_hold)) \
		{ \
			strcpy(games[num_games].parent_name, parent_hold); \
			num_clones++; \
		} \
		strcpy(games[num_games].romof, romof_hold); \
		strcpy(games[num_games].sampleof, sampleof_hold); \
		games[num_games].flags=flags_hold; \
		if (flags_hold & RESOURCE) \
			num_resources++; \
		num_games++; \
		processing_game=0; \
	} \
	parent_hold[0]='\0'; \
	romof_hold[0]='\0'; \
	sampleof_hold[0]='\0'; \
	flags_hold=0; \
}

/* --------------------------------------------------------------------------
 * HOLD macros
 *
 * Hold game/ROM details prior to them being stored
 * -------------------------------------------------------------------------- */

#define HOLD_GAME_FLAG(FLAGS) \
{ \
	flags_hold|=FLAGS; \
}

/* --- HOLD_GAME_TITLE will skip preceding spaces (for Callus) --- */

#define HOLD_GAME_TITLE(ST) \
{ \
	st_ptr=ST; \
	while (st_ptr[0]==' ') {st_ptr++;} \
	strncpy(title_hold, st_ptr, TITLE_LENGTH); \
	title_hold[TITLE_LENGTH]='\0'; \
}

/* --- HOLD_GAME_NAME will change to lower and decide whether to process  --- */

#define HOLD_GAME_NAME(ST) \
{ \
	strncpy(game_hold, ST, GAME_NAME_LENGTH); \
	game_hold[GAME_NAME_LENGTH]='\0'; \
	if (!game || \
			game && !strcmp(game_hold, game)) \
		processing_game=1; \
}

/* --- If game is a clone of specified game (-g option) and -c used... --- */

#define HOLD_PARENT_NAME(ST) \
{ \
	strncpy(parent_hold, ST, GAME_NAME_LENGTH); \
	parent_hold[GAME_NAME_LENGTH]='\0'; \
	if (!game || \
			game && !strcmp(game_hold, game) || \
			game && clones && !strcmp(parent_hold, game)) \
		processing_game=1; \
}

/* --- 'romof neogeo' means that the game is a Neo-Geo game. Set flag --- */

#define HOLD_ROMOF(ST) \
{ \
	strncpy(romof_hold, ST, GAME_NAME_LENGTH); \
	romof_hold[GAME_NAME_LENGTH]='\0'; \
	if (!strcmp(ST, "neogeo")) \
		flags_hold|=NEOGEO_GAME; \
}

#define HOLD_SAMPLEOF(ST) \
{ \
	strncpy(sampleof_hold, ST, GAME_NAME_LENGTH); \
	sampleof_hold[GAME_NAME_LENGTH]='\0'; \
}

/* --------------------------------------------------------------------------
 * STORE_ROM macro
 *
 * Works for name alone or full details on one line.
 * If current game/resource is being processed, the ROM is stored in memory.
 * STORE_ROM also clears out hold variables at the end.
 * -------------------------------------------------------------------------- */

#define GET_TOKEN(TOKEN, ST_PTR) \
{ \
	int token_idx=0; \
\
	while (ST_PTR[0]==' ' || ST_PTR[0]=='	') \
		ST_PTR++; \
\
	if (ST_PTR[0]=='"') \
	{ \
		ST_PTR++; \
		while (ST_PTR[0]!='"' && ST_PTR[0]!='\0') \
		{ \
			TOKEN[token_idx++]=ST_PTR[0]; \
			ST_PTR++; \
		} \
		TOKEN[token_idx]='\0';  \
		if (ST_PTR[0]!='\0') \
			ST_PTR++; \
	} \
	else \
	{ \
		while (ST_PTR[0]!=' ' && ST_PTR[0]!='	' && ST_PTR[0]!='\0') \
		{ \
			TOKEN[token_idx++]=ST_PTR[0]; \
			ST_PTR++; \
		} \
		TOKEN[token_idx]='\0';  \
	} \
}

#define STORE_ROM(ST) \
{ \
	if (processing_game) \
	{ \
		/* --- Link into game's list of ROMs --- */ \
		if (!games[num_games].first_rom) \
		{ \
			games[num_games].first_rom=&roms[num_roms]; \
			if (dat_info->flags & ALT_ROM_SUPPORT) \
				games[num_games].first_alt_rom=&alt_roms[num_roms]; \
		} \
		else \
		{ \
			roms[num_roms-1].next_rom=&roms[num_roms]; \
			if (dat_info->flags & ALT_ROM_SUPPORT) \
				alt_roms[num_roms-1].next_alt_rom=&alt_roms[num_roms]; \
		} \
\
		/* --- If it's not full ROM details, just store the name --- */ \
		if (strncmp(ST, "rom (", 5)) \
		{ \
			strncpy(roms[num_roms].name, ST, FILE_NAME_LENGTH); \
			roms[num_roms++].name[FILE_NAME_LENGTH]='\0'; \
		} \
		else \
		{ \
			/* --- Extract details and store them --- */ \
			st_ptr=ST; \
			GET_TOKEN(token, st_ptr) /* skip rom token */ \
			GET_TOKEN(token, st_ptr) /* skip bracket */ \
			GET_TOKEN(token, st_ptr) \
			alt_name=0; \
			while (token[0]!='\0') \
			{ \
				/* --- ROM name (can be several for dats with alt names) --- */ \
				if (!strcmp(token, "name")) \
				{ \
					GET_TOKEN(token, st_ptr) \
					strncpy(roms[num_roms].name, token, FILE_NAME_LENGTH); \
					roms[num_roms].name[FILE_NAME_LENGTH]='\0'; \
					if (dat_info->flags & ALT_ROM_SUPPORT) \
					{ \
						strncpy(alt_roms[num_roms].name[alt_name], token, FILE_NAME_LENGTH); \
						alt_roms[num_roms].name[alt_name][FILE_NAME_LENGTH]='\0'; \
						if (!(dat_info->flags & SIZE_SUPPORT)) /* otherwise the size will do it */ \
							alt_name++; \
					} \
					token[0]='\0'; \
				} \
				/* --- ROM 'merge' name --- */ \
				if (!strcmp(token, "merge")) \
				{ \
					GET_TOKEN(token, st_ptr) \
					strncpy(roms[num_roms].merge, token, FILE_NAME_LENGTH); \
					roms[num_roms].merge[FILE_NAME_LENGTH]='\0'; \
					token[0]='\0'; \
				} \
				/* --- ROM size --- */ \
				if (!strcmp(token, "size")) \
				{ \
					GET_TOKEN(token, st_ptr) \
					while (token[0]=='0' && token[0]!='\0') \
						strcpy(token, token+1); \
					strncpy(roms[num_roms].size, token, SIZE_LENGTH); \
					roms[num_roms].size[SIZE_LENGTH]='\0'; \
					if (dat_info->flags & ALT_ROM_SUPPORT) \
					{ \
						strncpy(alt_roms[num_roms].size[alt_name], token, SIZE_LENGTH); \
						alt_roms[num_roms].size[alt_name][SIZE_LENGTH]='\0'; \
						alt_name++; \
					} \
				} \
				/* --- ROM CRC --- */ \
				if (!strcmp(token, "crc") || !strcmp(token, "crc32")) \
				{ \
					GET_TOKEN(token, st_ptr) \
					sprintf(roms[num_roms].crc, "%08s", token); \
				} \
				/* --- Onto next property... --- */ \
				GET_TOKEN(token, st_ptr) \
			} \
			/* --- That's one more ROM (and possibly a resource ROM) --- */ \
			num_roms++; \
			if (flags_hold & RESOURCE) \
				num_resource_roms++; \
		} \
\
	} \
}

/* --- STORE_ROM_UNIQUE checks that ROM isn't already specified for game --- */
/* --- Used for Retrocade Romlist and System16 GCS files --- */

#define STORE_ROM_UNIQUE(ST) \
if (processing_game) \
{ \
	int found_dup=0; \
	rom_ptr=games[num_games].first_rom; \
	while (!found_dup && rom_ptr) \
	{ \
		if (!strcmp(ST, rom_ptr->name)) \
			found_dup++; \
		rom_ptr=rom_ptr->next_rom; \
	} \
	if (!found_dup) \
		STORE_ROM(ST) \
}

/* --- STORE_S16_ROMS can cope with how ROMs appear in GCS files --- */

#define STORE_S16_ROMS(ST, SKIP, MAX) \
{ \
	char *rom; \
	int i; \
	LOWER(ST) \
	rom=strtok(ST," "); \
	for (i=0; i<SKIP; i++) rom=strtok(NULL," "); \
\
	i=0; \
	while (i++<MAX && rom) \
	{ \
		STORE_ROM_UNIQUE(rom) \
		rom=strtok(NULL," "); \
	} \
}

/* --- Samples are straightforward --- */

#define STORE_SAMPLE(ST) \
{ \
	if (processing_game) \
	{ \
		if (!games[num_games].first_sample) \
			games[num_games].first_sample=&samples[num_samples]; \
		else \
			samples[num_samples-1].next_sample=&samples[num_samples]; \
\
		strncpy(samples[num_samples].name, ST, FILE_NAME_LENGTH); \
		samples[num_samples++].name[FILE_NAME_LENGTH]='\0'; \
	} \
}

/* --- DONE_LOAD does a bit of tidying up (such as identifying parents) --- */

#define DONE_LOAD \
{ \
	fclose(in); \
\
	dat_info->num_games=num_games; \
	dat_info->num_clones=num_clones; \
	dat_info->num_resources=num_resources; \
	dat_info->num_roms=num_roms; \
	dat_info->num_resource_roms=num_resource_roms; \
	dat_info->num_samples=num_samples; \
\
	return(identify_parents(dat_info, game)); \
}

